'''
Created on Mar 18, 2016

@author: nicolas
'''

import os
import sys
import re
import inspect
import collections

IS_PYTHON3 = sys.version_info > (3,)
IS_PYTHON2 = sys.version_info < (3,)


def humanize_duration(duration, show_milliseconds=False):
    ret = ""

    if duration / 3600 >= 1:
        ret += ("%02dh" if ret else "%dh") % (duration / 3600)
        duration %= 3600

    if duration / 60 >= 1:
        ret += ("%02dm" if ret else "%dm") % (duration / 60)
        duration %= 60

    if show_milliseconds:
        if duration >= 0:
            ret += ("%06.03fs" if ret else "%.03fs") % duration
    else:
        if duration >= 1:
            ret += ("%02ds" if ret else "%ds") % duration
        if ret == "":
            ret = "%.03fs" % duration

    return ret


def get_status_color(status):
    if status == "passed":
        return "green"
    elif status == "failed":
        return "red"
    elif status == "disabled":
        return "cyan"
    else:
        return "yellow"


def object_has_method(obj, method_name):
    try:
        return callable(getattr(obj, method_name))
    except AttributeError:
        return False


def get_callable_args(cb):
    spec = inspect.getfullargspec(cb) if IS_PYTHON3 else inspect.getargspec(cb)
    args_start = 1 if inspect.ismethod(cb) else 0
    return spec.args[args_start:]


def get_resource_path(relpath):
    return os.path.join(os.path.dirname(__file__), "resources", relpath)


def dict_cat(d1, d2):
    new = dict()
    new.update(d1)
    new.update(d2)
    return new


def is_string(s):
    if IS_PYTHON2:
        return type(s) in (str, unicode)
    else:
        return type(s) is str


# borrowed from https://stackoverflow.com/a/1176023
def camel_case_to_snake_case(name):
    s1 = re.sub('(.)([A-Z][a-z]+)', r'\1_\2', name)
    return re.sub('([a-z0-9])([A-Z])', r'\1_\2', s1).lower()


###
# OrderedSet, by Raymond Hettinger
# see http://code.activestate.com/recipes/576694-orderedset/
##

class OrderedSet(collections.MutableSet):

    def __init__(self, iterable=None):
        self.end = end = []
        end += [None, end, end]  # sentinel node for doubly linked list
        self.map = {}  # key --> [key, prev, next]
        if iterable is not None:
            self |= iterable

    def __len__(self):
        return len(self.map)

    def __contains__(self, key):
        return key in self.map

    def add(self, key):
        if key not in self.map:
            end = self.end
            curr = end[1]
            curr[2] = end[1] = self.map[key] = [key, curr, end]

    def discard(self, key):
        if key in self.map:
            key, prev, next = self.map.pop(key)
            prev[2] = next
            next[1] = prev

    def update(self, other):  # the update method is the only modification to the original class
        self |= other

    def __iter__(self):
        end = self.end
        curr = end[2]
        while curr is not end:
            yield curr[0]
            curr = curr[2]

    def __reversed__(self):
        end = self.end
        curr = end[1]
        while curr is not end:
            yield curr[0]
            curr = curr[1]

    def pop(self, last=True):
        if not self:
            raise KeyError('set is empty')
        key = self.end[1][0] if last else self.end[2][0]
        self.discard(key)
        return key

    def __repr__(self):
        if not self:
            return '%s()' % (self.__class__.__name__,)
        return '%s(%r)' % (self.__class__.__name__, list(self))

    def __eq__(self, other):
        if isinstance(other, OrderedSet):
            return len(self) == len(other) and list(self) == list(other)
        return set(self) == set(other)



import sys
from multiprocessing.dummy import Pool, Queue

from lemoncheesecake.exceptions import AbortTask, TasksExecutionFailure, CircularDependencyError, \
    serialize_current_exception


class BaseTask(object):
    def __init__(self):
        self.successful = None
        self.exception = None

    def get_dependencies(self):
        return set()

    def run(self, context):
        pass

    def abort(self, context):
        pass


def pop_runnable_tasks(remaining_tasks, completed_tasks, nb_tasks):
    runnable_tasks = [task for task in remaining_tasks if set(task.get_dependencies()).issubset(completed_tasks)]
    runnable_tasks_by_priority = sorted(runnable_tasks, key=lambda task: len(task.get_dependencies()), reverse=True)

    for task in runnable_tasks_by_priority[:nb_tasks]:
        remaining_tasks.remove(task)
        yield task


def pop_dependant_tasks(ref_task, tasks):
    for task in list(tasks):
        if ref_task in task.get_dependencies():
            tasks.remove(task)
            yield task
            for indirect_task in pop_dependant_tasks(task, tasks):
                yield indirect_task


def run_task(task, context, completed_task_queue):
    try:
        task.run(context)
    except AbortTask:
        task.successful = False
    except Exception:
        task.successful = False
        task.exception = serialize_current_exception()
    else:
        task.successful = True

    completed_task_queue.put(task)


def do_run_tasks(tasks, context, pool, completed_tasks_queue):
    for task in tasks:
        pool.apply_async(run_task, args=(task, context, completed_tasks_queue))


def abort_task(task, context, completed_task_queue):
    try:
        task.abort(context)
    except Exception:
        task.successful = False
        task.exception = serialize_current_exception()
    else:
        task.successful = True

    completed_task_queue.put(task)


def abort_tasks(tasks, context, pool, completed_tasks_queue):
    for task in tasks:
        pool.apply_async(abort_task, args=(task, context, completed_tasks_queue))


def abort_all_tasks(tasks, remaining_tasks, completed_tasks, context, pool, completed_tasks_queue):
    abort_tasks(remaining_tasks, context, pool, completed_tasks_queue)
    while len(completed_tasks) != len(tasks):
        completed_task = completed_tasks_queue.get()
        completed_tasks.append(completed_task)


def run_tasks(tasks, context=None, nb_threads=1, watchdog=None):
    for task in tasks:
        check_task_dependencies(task)

    remaining_tasks = list(tasks)
    completed_tasks = list()

    pool = Pool(nb_threads)
    completed_tasks_queue = Queue()

    try:
        do_run_tasks(pop_runnable_tasks(remaining_tasks, completed_tasks, nb_threads), context, pool, completed_tasks_queue)

        while len(completed_tasks) != len(tasks):
            # wait for one task to complete
            completed_task = completed_tasks_queue.get()
            completed_tasks.append(completed_task)

            # watchdog
            exception_class, exception_arg = watchdog() if watchdog else (None, None)
            if exception_class:
                abort_all_tasks(tasks, remaining_tasks, completed_tasks, context, pool, completed_tasks_queue)
                raise exception_class(exception_arg)

            # schedule next tasks depending on the completed task success
            if completed_task.successful:
                runnable_tasks = pop_runnable_tasks(remaining_tasks, completed_tasks, nb_threads)
                do_run_tasks(runnable_tasks, context, pool, completed_tasks_queue)
            else:
                abortable_tasks = pop_dependant_tasks(completed_task, remaining_tasks)
                abort_tasks(abortable_tasks, context, pool, completed_tasks_queue)

    except KeyboardInterrupt:
        abort_all_tasks(tasks, remaining_tasks, completed_tasks, context, pool, completed_tasks_queue)

    finally:
        pool.close()

    if any(task.exception for task in tasks):
        raise TasksExecutionFailure(
            "Caught exceptions:\n%s" % "\n".join(task.exception for task in tasks if task.exception)
        )


def check_task_dependencies(task, task_path=()):
    for task_in_path in task_path:
        if task_in_path in task.get_dependencies():
            raise CircularDependencyError("Task %s has a circular dependency on task %s" % (task, task_in_path))

    for dependency in task.get_dependencies():
        check_task_dependencies(dependency, task_path=(task,) + task_path)



'''
Created on Jan 24, 2016

@author: nicolas
'''

import os
import sys
import traceback
import threading
import itertools

from lemoncheesecake.utils import IS_PYTHON3
from lemoncheesecake.runtime import *
from lemoncheesecake.runtime import initialize_runtime, set_runtime_location, is_location_successful
from lemoncheesecake.reporting import Report, initialize_report_writer, initialize_reporting_backends
from lemoncheesecake.exceptions import AbortTest, AbortSuite, AbortAllTests, FixtureError, \
    UserError, serialize_current_exception
from lemoncheesecake import events
from lemoncheesecake.testtree import TreeLocation, flatten_tests
from lemoncheesecake.task import BaseTask, run_tasks


class RunContext(object):
    def __init__(self, stop_on_failure, fixture_registry):
        self.stop_on_failure = stop_on_failure
        self.fixture_registry = fixture_registry
        self._abort_tests = False
        self._aborted_suites = set()

    def abort_suite(self, suite):
        self._aborted_suites.add(suite)

    def is_suite_aborted(self, suite):
        return suite in self._aborted_suites

    def abort_session(self):
        self._abort_tests = True

    def is_session_aborted(self):
        return self._abort_tests

    def handle_exception(self, excp, suite=None):
        if isinstance(excp, AbortTest):
            log_error(str(excp))
        elif isinstance(excp, AbortSuite):
            log_error(str(excp))
            self.abort_suite(suite)
        elif isinstance(excp, AbortAllTests):
            log_error(str(excp))
            self.abort_session()
        else:
            # FIXME: use exception instead of last implicit stacktrace
            stacktrace = traceback.format_exc()
            if not IS_PYTHON3:
                stacktrace = stacktrace.decode("utf-8", "replace")
            log_error("Caught unexpected exception while running test: " + stacktrace)

    def run_setup_funcs(self, funcs, location):
        teardown_funcs = []
        for setup_func, teardown_func in funcs:
            if setup_func:
                try:
                    setup_func()
                except Exception as e:
                    self.handle_exception(e)
                    break
                else:
                    if not is_location_successful(location):
                        break
                    else:
                        teardown_funcs.append(teardown_func)
            else:
                teardown_funcs.append(teardown_func)
        return teardown_funcs

    def run_teardown_funcs(self, teardown_funcs):
        count = 0
        for teardown_func in teardown_funcs:
            if teardown_func:
                try:
                    teardown_func()
                except Exception as e:
                    self.handle_exception(e)
                else:
                    count += 1
            else:
                count += 1
        return count


class TestTask(BaseTask):
    def __init__(self, test, suite_scheduled_fixtures, dependency=None):
        BaseTask.__init__(self)
        self.test = test
        self.suite_scheduled_fixtures = suite_scheduled_fixtures
        self.dependency = dependency

    def get_dependencies(self):
        return (self.dependency,) if self.dependency else ()

    def abort(self, context, reason=""):
        events.fire(events.TestSkippedEvent(self.test, reason))

    def run(self, context):
        suite = self.test.parent_suite

        ###
        # Checker whether the test must be executed or not
        ###
        if self.test.is_disabled():
            events.fire(events.TestDisabledEvent(self.test, ""))
            return

        if context.is_suite_aborted(suite):
            self.abort(context, "Cannot execute this test: the tests of this test suite have been aborted.")
            return

        if context.is_session_aborted():
            self.abort(context, "Cannot execute this test: all tests have been aborted.")
            return

        ###
        # Begin test
        ###
        events.fire(events.TestStartEvent(self.test))
        set_runtime_location(TreeLocation.in_test(self.test))

        ###
        # Setup test (setup and fixtures)
        ###
        test_setup_error = False
        setup_teardown_funcs = list()

        setup_teardown_funcs.append([
            (lambda: suite.get_hook("setup_test")(self.test.name)) if suite.has_hook("setup_test") else None,
            (lambda: suite.get_hook("teardown_test")(self.test.name)) if suite.has_hook("teardown_test") else None
        ])
        scheduled_fixtures = context.fixture_registry.get_fixtures_scheduled_for_test(
            self.test, self.suite_scheduled_fixtures
        )
        setup_teardown_funcs.extend(scheduled_fixtures.get_setup_teardown_pairs())

        if any(setup for setup, _ in setup_teardown_funcs):
            events.fire(events.TestSetupStartEvent(self.test))
            set_step("Setup test")
            teardown_funcs = context.run_setup_funcs(setup_teardown_funcs, TreeLocation.in_test(self.test))
            if len(teardown_funcs) != len(setup_teardown_funcs):
                test_setup_error = True
            events.fire(events.TestSetupEndEvent(self.test, not test_setup_error))
        else:
            teardown_funcs = [teardown for _, teardown in setup_teardown_funcs if teardown]

        if context.stop_on_failure and test_setup_error:
            context.abort_session()

        ###
        # Run test:
        ###
        if not test_setup_error:
            test_func_params = scheduled_fixtures.get_fixture_results(self.test.get_fixtures())
            set_step(self.test.description)
            try:
                self.test.callback(**test_func_params)
            except Exception as e:
                context.handle_exception(e, suite)

        ###
        # Teardown
        ###
        if any(teardown_funcs):
            events.fire(events.TestTeardownStartEvent(self.test))
            set_step("Teardown test")
            context.run_teardown_funcs(teardown_funcs)
            events.fire(events.TestTeardownEndEvent(self.test, is_location_successful(TreeLocation.in_test(self.test))))

        if context.stop_on_failure and not is_location_successful(TreeLocation.in_test(self.test)):
            context.abort_session()

        events.fire(events.TestEndEvent(self.test))

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.test.path)


def build_test_task(test, suite_scheduled_fixtures, dependency):
    return TestTask(test, suite_scheduled_fixtures, dependency)


def build_suite_tasks(
        suite, fixture_registry, session_scheduled_fixtures, test_session_setup_task,
        parent_suite_beginning_task=None):
    ###
    # Build suite beginning task
    ###
    suite_beginning_task = build_suite_beginning_task(
        suite, list(filter(bool, (test_session_setup_task, parent_suite_beginning_task)))
    )

    ###
    # Build suite setup task (if any)
    ###
    suite_scheduled_fixtures = fixture_registry.get_fixtures_scheduled_for_suite(
        suite, session_scheduled_fixtures
    )
    suite_setup_task = build_suite_initialization_task(suite, suite_scheduled_fixtures, (suite_beginning_task,))

    ###
    # Build test tasks
    ###
    test_dependency = suite_setup_task if suite_setup_task else suite_beginning_task
    test_tasks = [
        build_test_task(test, suite_scheduled_fixtures, test_dependency)
        for test in suite.get_tests()
    ]

    ###
    # Build suite teardown task (if any)
    ###
    suite_teardown_task = build_suite_teardown_task(suite, suite_setup_task, test_tasks)

    ###
    # Build sub suite tasks
    ###
    sub_suite_tasks = []
    for sub_suite in suite.get_suites():
        sub_suite_tasks.extend(
            build_suite_tasks(
                sub_suite, fixture_registry, session_scheduled_fixtures, test_session_setup_task, suite_beginning_task
            )
        )

    ###
    # Build suite ending task
    ###
    suite_ending_dependencies = []
    suite_ending_dependencies.extend(test_tasks)
    if suite_teardown_task:
        suite_ending_dependencies.append(suite_teardown_task)
    suite_ending_dependencies.extend(
        task for task in sub_suite_tasks if isinstance(task, SuiteEndingTask) and task.suite in suite.get_suites()
    )
    suite_ending_task = build_suite_ending_task(suite, suite_ending_dependencies)

    ###
    # Return tasks != None
    ###
    task_iter = itertools.chain(
        (suite_beginning_task,),
        (suite_setup_task,), test_tasks, (suite_teardown_task,), sub_suite_tasks,
        (suite_ending_task,)
    )
    return list(filter(bool, task_iter))


class SuiteBeginningTask(BaseTask):
    def __init__(self, suite, dependencies):
        BaseTask.__init__(self)
        self.suite = suite
        self._dependencies = dependencies

    def get_dependencies(self):
        return self._dependencies

    def run(self, context):
        events.fire(events.SuiteStartEvent(self.suite))

    abort = run

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.suite.path)


def build_suite_beginning_task(suite, dependencies):
    return SuiteBeginningTask(suite, dependencies)


class SuiteInitializationTask(BaseTask):
    def __init__(self, suite, setup_teardown_funcs, dependencies):
        BaseTask.__init__(self)
        self.suite = suite
        self.setup_teardown_funcs = setup_teardown_funcs
        self._dependencies = dependencies
        self.teardown_funcs = []

    def get_dependencies(self):
        return self._dependencies

    @staticmethod
    def begin_suite_setup(suite):
        events.fire(events.SuiteSetupStartEvent(suite))
        set_runtime_location(TreeLocation.in_suite_setup(suite))
        set_step("Setup suite")

    @staticmethod
    def end_suite_setup(suite):
        events.fire(events.SuiteSetupEndEvent(suite))

    def run(self, context):
        if context.is_session_aborted():
            return

        if any(setup for setup, _ in self.setup_teardown_funcs):
            SuiteInitializationTask.begin_suite_setup(self.suite)
            self.teardown_funcs = context.run_setup_funcs(
                self.setup_teardown_funcs, TreeLocation.in_suite_setup(self.suite)
            )
            if len(self.teardown_funcs) != len(self.setup_teardown_funcs):
                context.abort_suite(self.suite)
            SuiteInitializationTask.end_suite_setup(self.suite)
        else:
            self.teardown_funcs = [teardown for _, teardown in self.setup_teardown_funcs if teardown]

        if context.stop_on_failure and context.is_suite_aborted(self.suite):
            context.abort_session()

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.suite.path)


def wrap_setup_suite(suite, scheduled_fixtures):
    setup_suite = suite.get_hook("setup_suite")
    if setup_suite is None:
        return None

    fixtures_names = suite.get_hook_params("setup_suite")
    def wrapper():
        fixtures = scheduled_fixtures.get_fixture_results(fixtures_names)
        setup_suite(**fixtures)

    return wrapper


def build_suite_initialization_task(suite, scheduled_fixtures, dependencies):
    setup_teardown_funcs = []

    if not scheduled_fixtures.is_empty():
        setup_teardown_funcs.extend(scheduled_fixtures.get_setup_teardown_pairs())

    if suite.get_injected_fixture_names():
        setup_teardown_funcs.append((
            lambda: suite.inject_fixtures(scheduled_fixtures.get_fixture_results(suite.get_injected_fixture_names())),
            None
        ))

    if suite.has_hook("setup_suite") or suite.has_hook("teardown_suite"):
        setup_teardown_funcs.append([
            wrap_setup_suite(suite, scheduled_fixtures),
            suite.get_hook("teardown_suite")
        ])

    return SuiteInitializationTask(suite, setup_teardown_funcs, dependencies) if setup_teardown_funcs else None


class SuiteEndingTask(BaseTask):
    def __init__(self, suite, dependencies):
        BaseTask.__init__(self)
        self.suite = suite
        self._dependencies = dependencies

    def get_dependencies(self):
        return self._dependencies

    def run(self, context):
        events.fire(events.SuiteEndEvent(self.suite))

    abort = run

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.suite.path)


def build_suite_ending_task(suite, dependencies):
    return SuiteEndingTask(suite, dependencies)


class SuiteTeardownTask(BaseTask):
    def __init__(self, suite, suite_setup_task, dependencies):
        BaseTask.__init__(self)
        self.suite = suite
        self.suite_setup_task = suite_setup_task
        self._dependencies = dependencies

    def get_dependencies(self):
        return self._dependencies

    @staticmethod
    def begin_suite_teardown(suite):
        events.fire(events.SuiteTeardownStartEvent(suite))
        set_runtime_location(TreeLocation.in_suite_teardown(suite))
        set_step("Teardown suite")

    @staticmethod
    def end_suite_teardown(suite):
        events.fire(events.SuiteTeardownEndEvent(suite))

    def run(self, context):
        if any(self.suite_setup_task.teardown_funcs):
            SuiteTeardownTask.begin_suite_teardown(self.suite)
            context.run_teardown_funcs(self.suite_setup_task.teardown_funcs)
            if context.stop_on_failure and not is_location_successful(TreeLocation.in_suite_teardown(self.suite)):
                context.abort_session()
            SuiteTeardownTask.end_suite_teardown(self.suite)

    def __str__(self):
        return "<%s %s>" % (self.__class__.__name__, self.suite.path)


def build_suite_teardown_task(suite, suite_setup_task, dependencies):
    return SuiteTeardownTask(suite, suite_setup_task, dependencies) if suite_setup_task else None


class TestSessionSetupTask(BaseTask):
    def __init__(self, scheduled_fixtures):
        BaseTask.__init__(self)
        self.scheduled_fixtures = scheduled_fixtures
        self.teardown_funcs = []

    @staticmethod
    def begin_test_session_setup():
        events.fire(events.TestSessionSetupStartEvent())
        set_runtime_location(TreeLocation.in_test_session_setup())
        set_step("Setup test session")

    @staticmethod
    def end_test_session_setup():
        events.fire(events.TestSessionSetupEndEvent())

    def run(self, context):
        setup_teardown_funcs = self.scheduled_fixtures.get_setup_teardown_pairs()

        if any(setup for setup, _ in setup_teardown_funcs):
            TestSessionSetupTask.begin_test_session_setup()
            self.teardown_funcs = context.run_setup_funcs(setup_teardown_funcs, TreeLocation.in_test_session_setup())
            if len(self.teardown_funcs) != len(setup_teardown_funcs):
                context.abort_session()
            TestSessionSetupTask.end_test_session_setup()
        else:
            self.teardown_funcs = [teardown for _, teardown in setup_teardown_funcs if teardown]


def build_test_session_setup_task(scheduled_fixtures):
    return TestSessionSetupTask(scheduled_fixtures) if not scheduled_fixtures.is_empty() else None


class TestSessionTeardownTask(BaseTask):
    def __init__(self, test_session_setup_task, dependencies):
        BaseTask.__init__(self)
        self.test_session_setup_task = test_session_setup_task
        self._dependencies = dependencies

    def get_dependencies(self):
        return self._dependencies

    @staticmethod
    def begin_test_session_teardown():
        events.fire(events.TestSessionTeardownStartEvent())
        set_runtime_location(TreeLocation.in_test_session_teardown())
        set_step("Teardown test session")

    @staticmethod
    def end_test_session_teardown():
        events.fire(events.TestSessionTeardownEndEvent())

    def run(self, context):
        if any(self.test_session_setup_task.teardown_funcs):
            TestSessionTeardownTask.begin_test_session_teardown()
            context.run_teardown_funcs(self.test_session_setup_task.teardown_funcs)
            TestSessionTeardownTask.end_test_session_teardown()


def build_test_session_teardown_task(test_session_setup_task, dependencies):
    return TestSessionTeardownTask(test_session_setup_task, dependencies) if test_session_setup_task else None


def build_tasks(suites, fixture_registry, session_scheduled_fixtures):
    ###
    # Build test session setup task
    ###
    test_session_setup_task = build_test_session_setup_task(session_scheduled_fixtures)

    ###
    # Build suite tasks
    ###
    suite_tasks = []
    for suite in suites:
        suite_tasks.extend(
            build_suite_tasks(suite, fixture_registry, session_scheduled_fixtures, test_session_setup_task)
        )

    ###
    # Build test session teardown task
    ###
    if test_session_setup_task:
        test_session_teardown_dependencies = [
            task for task in suite_tasks if isinstance(task, SuiteEndingTask) and task.suite in suites
        ]
        test_session_teardown_task = build_test_session_teardown_task(
            test_session_setup_task, test_session_teardown_dependencies
        )
    else:
        test_session_teardown_task = None

    ###
    # Return tasks != None
    ###
    task_iter = itertools.chain((test_session_setup_task,), suite_tasks, (test_session_teardown_task,))
    return list(filter(bool, task_iter))


def watchdog():
    exception, serialized_exception = events.get_pending_failure()
    if exception:
        return exception.__class__, serialized_exception
    else:
        return None, None


def run_session(suites, fixture_registry, prerun_session_scheduled_fixtures, reporting_backends, report_dir,
                stop_on_failure=False, nb_threads=1):
    # initialize runtime & global test variables
    report = Report()
    report.nb_threads = nb_threads
    report.add_info("Command line", " ".join([os.path.basename(sys.argv[0])] + sys.argv[1:]))
    session = initialize_report_writer(report)
    nb_tests = len(list(flatten_tests(suites)))
    initialize_runtime(report_dir, report, prerun_session_scheduled_fixtures)
    initialize_reporting_backends(
        reporting_backends, report_dir, report, parallel=nb_threads > 1 and nb_tests > 1
    )

    # build tasks and run context
    session_scheduled_fixtures = fixture_registry.get_fixtures_scheduled_for_session(
        suites, prerun_session_scheduled_fixtures
    )
    tasks = build_tasks(suites, fixture_registry, session_scheduled_fixtures)
    context = RunContext(stop_on_failure, fixture_registry)

    # start event handler thread
    event_handler_thread = threading.Thread(target=events.handler_loop)
    event_handler_thread.start()

    try:
        events.fire(events.TestSessionStartEvent(report))
        run_tasks(tasks, context, nb_threads, watchdog)
        events.fire(events.TestSessionEndEvent(report))
    finally:
        # wait for event handler to finish
        events.end_of_events()
        event_handler_thread.join()

    exception, serialized_exception = events.get_pending_failure()
    if exception:
        raise exception.__class__(serialized_exception)

    return report


def run_suites(suites, fixture_registry, reporting_backends, report_dir, stop_on_failure=False, nb_threads=1):
    fixture_teardowns = []

    # setup pre_session fixtures
    errors = []
    scheduled_fixtures = fixture_registry.get_fixtures_scheduled_for_session_prerun(suites)
    for setup, teardown in scheduled_fixtures.get_setup_teardown_pairs():
        try:
            setup()
        except UserError:
            raise
        except Exception:
            errors.append("Got the following exception when executing fixture (scope 'session_prerun')%s" % (
                serialize_current_exception(show_stacktrace=True)
            ))
            break
        fixture_teardowns.append(teardown)

    if not errors:
        report = run_session(
            suites, fixture_registry, scheduled_fixtures,
            reporting_backends, report_dir, stop_on_failure=stop_on_failure, nb_threads=nb_threads
        )
    else:
        report = None

    # teardown pre_session fixtures
    for teardown in fixture_teardowns:
        try:
            teardown()
        except UserError:
            raise
        except Exception:
            errors.append("Got the following exception on fixture teardown (scope 'session_prerun')%s" % (
                serialize_current_exception(show_stacktrace=True)
            ))

    if errors:
        raise FixtureError("\n".join(errors))

    return report.is_successful() if report else False



'''
Created on Apr 4, 2017

@author: nicolas
'''

import os
import re
import difflib
import json


from lemoncheesecake.matching.base import MatchExpected, match_result, match_success, match_failure, got_value, to_be

__all__ = (
    "starts_with", "ends_with", "contains_string", "match_pattern", "is_text", "is_json"
)

_REGEXP_TYPE = type(re.compile("dummy"))


class StartsWith(MatchExpected):
    def description(self, conjugate=False):
        return '%s with "%s"' % ("starts" if conjugate else "to start", self.expected)

    def matches(self, actual):
        return match_result(actual.startswith(self.expected), got_value(actual))


def starts_with(s):
    """Test if string begins with given prefix"""
    return StartsWith(s)


class EndsWith(MatchExpected):
    def description(self, conjugate=False):
        return '%s with "%s"' % ("ends" if conjugate else "to end", self.expected)

    def matches(self, actual):
        return match_result(actual.endswith(self.expected), got_value(actual))


def ends_with(s):
    """Test if string ends with given suffix"""
    return EndsWith(s)


class ContainsString(MatchExpected):
    def description(self, conjugate=False):
        return '%s "%s"' % ("contains" if conjugate else "to contain", self.expected)

    def matches(self, actual):
        return match_result(self.expected in actual, got_value(actual))


def contains_string(s):
    """Test if string contains sub string"""
    return ContainsString(s)


class MatchPattern(MatchExpected):
    def description(self, conjugate=False):
        return '%s "%s"' % ("matches pattern" if conjugate else "to match pattern", self.expected.pattern)

    def matches(self, actual):
        return match_result(self.expected.search(actual) is not None, got_value(actual))


def match_pattern(pattern):
    """Test if string matches given pattern (using the `search` method of the `re` module)"""
    return MatchPattern(pattern if type(pattern) == _REGEXP_TYPE else re.compile(pattern))


def _diff_text(text1, text2, linesep):
    return "\n".join(
        difflib.unified_diff(
            text1.split(linesep), text2.split(linesep), fromfile="expected", tofile="actual", lineterm=""
        )
    )


class IsText(MatchExpected):
    def __init__(self, expected, linesep):
        MatchExpected.__init__(self, expected)
        self.linesep = linesep

    def description(self, conjugate=False):
        return "%s the text:\n<<<\n%s\n>>>\n" % (to_be(conjugate), self.expected)

    def matches(self, actual):
        if actual == self.expected:
            return match_success()
        else:
            return match_failure(
                "Text does not match:\n%s" % _diff_text(self.expected, actual, self.linesep)
            )


def is_text(expected, linesep=os.linesep):
    return IsText(expected, linesep)


def _format_json(data):
    return json.dumps(data, ensure_ascii=False, sort_keys=True, indent=4)


class IsJson(MatchExpected):
    def description(self, conjugate=False):
        return "%s JSON:\n%s\n" % (to_be(conjugate), _format_json(self.expected))

    def matches(self, actual):
        if actual == self.expected:
            return match_success()
        else:
            return match_failure(
                "JSON does not match:\n%s" % _diff_text(_format_json(self.expected), _format_json(actual), "\n")
            )


def is_json(expected):
    return IsJson(expected)



#!/usr/bin/python

'''
Created on Aug 20, 2016

@author: nicolas
'''

from setuptools import setup, find_packages

import lemoncheesecake


setup(
    name="lemoncheesecake",
    version=lemoncheesecake.__version__,
    description="A framework for functional QA testing",
    long_description=open("README.rst").read(),
    author="Nicolas Delon",
    author_email="nicolas.delon@gmail.com",
    license="Apache License (Version 2.0)",
    url="http://lemoncheesecake.io",
    classifiers=[
        "Development Status :: 4 - Beta",
        "Environment :: Console",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Software Development :: Quality Assurance",
        "Topic :: Software Development :: Testing",
    ],
    keywords="QA tests functional",

    packages=find_packages(),
    include_package_data=True,
    install_requires=["colorama", "termcolor", "terminaltables"],
    extras_require={
        "xml": "lxml",
        "junit": "lxml",
        "reportportal": "reportportal-client",
        "slack": "slacker"
    },
    entry_points={
        "console_scripts": [
            "lcc = lemoncheesecake.cli:main",
        ]
    }
)
